@@ -3,7 +3,7 @@
 //! * Only emits a single `Rename` event if the rename `From` and `To` events can be matched
 //! * Merges multiple `Rename` events
 //! * Takes `Rename` events into account and updates paths for events that occurred before the rename event, but which haven't been emitted, yet
-//! * Optionally keeps track of the file system IDs all files and stiches rename events together (FSevents, Windows)
+//! * Optionally keeps track of the file system IDs all files and stitches rename events together (macOS FS Events, Windows)
 //! * Emits only one `Remove` event when deleting a directory (inotify)
 //! * Doesn't emit duplicate create events
 //! * Doesn't emit `Modify` events after a `Create` event
@@ -12,7 +12,7 @@
 //!
 //! ```toml
 //! [dependencies]
-//! notify-debouncer-full = "0.3.1"
+//! notify-debouncer-full = "0.5.0"
 //! ```
 //!
 //! In case you want to select specific features of notify,
@@ -20,10 +20,10 @@
 //! Otherwise you can just use the re-export of notify from debouncer-full.
 //!
 //! ```toml
-//! notify-debouncer-full = "0.3.1"
+//! notify-debouncer-full = "0.5.0"
 //! notify = { version = "..", features = [".."] }
 //! ```
-//!  
+//!
 //! # Examples
 //!
 //! ```rust,no_run
@@ -42,60 +42,54 @@
 //!
 //! // Add a path to be watched. All files and directories at that path and
 //! // below will be monitored for changes.
-//! debouncer.watcher().watch(Path::new("."), RecursiveMode::Recursive).unwrap();
-//!
-//! // Add the same path to the file ID cache. The cache uses unique file IDs
-//! // provided by the file system and is used to stich together rename events
-//! // in case the notification back-end doesn't emit rename cookies.
-//! debouncer.cache().add_root(Path::new("."), RecursiveMode::Recursive);
+//! debouncer.watch(".", RecursiveMode::Recursive).unwrap();
 //! ```
 //!
 //! # Features
 //!
 //! The following crate features can be turned on or off in your cargo dependency config:
 //!
-//! - `crossbeam` enabled by default, adds [`DebounceEventHandler`](DebounceEventHandler) support for crossbeam channels.
-//!   Also enables crossbeam-channel in the re-exported notify. You may want to disable this when using the tokio async runtime.
-//! - `serde` enables serde support for events.
-//! 
+//! - `serde` passed down to notify-types, off by default
+//! - `web-time` passed down to notify-types, off by default
+//! - `crossbeam-channel` passed down to notify, off by default
+//! - `macos_fsevent` passed down to notify, off by default
+//! - `macos_kqueue` passed down to notify, off by default
+//! - `serialization-compat-6` passed down to notify, off by default
+//!
 //! # Caveats
-//! 
+//!
 //! As all file events are sourced from notify, the [known problems](https://docs.rs/notify/latest/notify/#known-problems) section applies here too.
 
 mod cache;
-mod debounced_event;
+mod time;
 
 #[cfg(test)]
 mod testing;
 
 use std::{
-    collections::{HashMap, VecDeque},
-    path::PathBuf,
+    cmp::Reverse,
+    collections::{BinaryHeap, HashMap, VecDeque},
+    path::{Path, PathBuf},
     sync::{
         atomic::{AtomicBool, Ordering},
-        Arc,
+        Arc, Mutex,
     },
-    time::Duration,
+    time::{Duration, Instant},
 };
 
-pub use cache::{FileIdCache, FileIdMap, NoCache};
-pub use debounced_event::DebouncedEvent;
+use time::now;
+
+pub use cache::{FileIdCache, FileIdMap, NoCache, RecommendedCache};
 
 pub use file_id;
 pub use notify;
+pub use notify_types::debouncer_full::DebouncedEvent;
 
 use file_id::FileId;
 use notify::{
     event::{ModifyKind, RemoveKind, RenameMode},
-    Error, ErrorKind, Event, EventKind, RecommendedWatcher, Watcher,
+    Error, ErrorKind, Event, EventKind, RecommendedWatcher, RecursiveMode, Watcher, WatcherKind,
 };
-use parking_lot::{MappedMutexGuard, Mutex, MutexGuard};
-
-#[cfg(test)]
-use mock_instant::Instant;
-
-#[cfg(not(test))]
-use std::time::Instant;
 
 /// The set of requirements for watcher debounce event handling functions.
 ///
@@ -131,7 +125,7 @@ where
     }
 }
 
-#[cfg(feature = "crossbeam")]
+#[cfg(feature = "crossbeam-channel")]
 impl DebounceEventHandler for crossbeam_channel::Sender<DebounceEventResult> {
     fn handle_event(&mut self, event: DebounceEventResult) {
         let _ = self.send(event);
@@ -182,6 +176,7 @@ impl Queue {
 #[derive(Debug)]
 pub(crate) struct DebounceDataInner<T> {
     queues: HashMap<PathBuf, Queue>,
+    roots: Vec<(PathBuf, RecursiveMode)>,
     cache: T,
     rename_event: Option<(DebouncedEvent, Option<FileId>)>,
     rescan_event: Option<DebouncedEvent>,
@@ -193,6 +188,7 @@ impl<T: FileIdCache> DebounceDataInner<T> {
     pub(crate) fn new(cache: T, timeout: Duration) -> Self {
         Self {
             queues: HashMap::new(),
+            roots: Vec::new(),
             cache,
             rename_event: None,
             rescan_event: None,
@@ -203,7 +199,7 @@ impl<T: FileIdCache> DebounceDataInner<T> {
 
     /// Retrieve a vec of debounced events, removing them if not continuous
     pub fn debounced_events(&mut self) -> Vec<DebouncedEvent> {
-        let now = Instant::now();
+        let now = now();
         let mut events_expired = Vec::with_capacity(self.queues.len());
         let mut queues_remaining = HashMap::with_capacity(self.queues.len());
 
@@ -216,6 +212,7 @@ impl<T: FileIdCache> DebounceDataInner<T> {
             }
         }
 
+        // drain the entire queue, then process the expired events and re-add the rest
         // TODO: perfect fit for drain_filter https://github.com/rust-lang/rust/issues/59618
         for (path, mut queue) in self.queues.drain() {
             let mut kind_index = HashMap::new();
@@ -249,28 +246,18 @@ impl<T: FileIdCache> DebounceDataInner<T> {
 
         self.queues = queues_remaining;
 
-        // order events for different files chronologically, but keep the order of events for the same file
-        events_expired.sort_by(|event_a, event_b| {
-            // use the last path because rename events are emitted for the target path
-            if event_a.paths.last() == event_b.paths.last() {
-                std::cmp::Ordering::Equal
-            } else {
-                event_a.time.cmp(&event_b.time)
-            }
-        });
-
-        events_expired
+        sort_events(events_expired)
     }
 
     /// Returns all currently stored errors
     pub fn errors(&mut self) -> Vec<Error> {
-        let mut v = Vec::new();
-        std::mem::swap(&mut v, &mut self.errors);
-        v
+        std::mem::take(&mut self.errors)
     }
 
     /// Add an error entry to re-send later on
     pub fn add_error(&mut self, error: Error) {
+        log::trace!("raw error: {error:?}");
+
         self.errors.push(error);
     }
 
@@ -279,8 +266,8 @@ impl<T: FileIdCache> DebounceDataInner<T> {
         log::trace!("raw event: {event:?}");
 
         if event.need_rescan() {
-            self.cache.rescan();
-            self.rescan_event = Some(event.into());
+            self.cache.rescan(&self.roots);
+            self.rescan_event = Some(DebouncedEvent { event, time: now() });
             return;
         }
 
@@ -288,9 +275,11 @@ impl<T: FileIdCache> DebounceDataInner<T> {
 
         match &event.kind {
             EventKind::Create(_) => {
-                self.cache.add_path(path);
+                let recursive_mode = self.recursive_mode(path);
 
-                self.push_event(event, Instant::now());
+                self.cache.add_path(path, recursive_mode);
+
+                self.push_event(event, now());
             }
             EventKind::Modify(ModifyKind::Name(rename_mode)) => {
                 match rename_mode {
@@ -316,23 +305,38 @@ impl<T: FileIdCache> DebounceDataInner<T> {
                 }
             }
             EventKind::Remove(_) => {
-                self.push_remove_event(event, Instant::now());
+                self.push_remove_event(event, now());
             }
             EventKind::Other => {
                 // ignore meta events
             }
             _ => {
                 if self.cache.cached_file_id(path).is_none() {
-                    self.cache.add_path(path);
+                    let recursive_mode = self.recursive_mode(path);
+
+                    self.cache.add_path(path, recursive_mode);
                 }
 
-                self.push_event(event, Instant::now());
+                self.push_event(event, now());
             }
         }
     }
 
+    fn recursive_mode(&mut self, path: &Path) -> RecursiveMode {
+        self.roots
+            .iter()
+            .find_map(|(root, recursive_mode)| {
+                if path.starts_with(root) {
+                    Some(*recursive_mode)
+                } else {
+                    None
+                }
+            })
+            .unwrap_or(RecursiveMode::NonRecursive)
+    }
+
     fn handle_rename_from(&mut self, event: Event) {
-        let time = Instant::now();
+        let time = now();
         let path = &event.paths[0];
 
         // store event
@@ -345,7 +349,9 @@ impl<T: FileIdCache> DebounceDataInner<T> {
     }
 
     fn handle_rename_to(&mut self, event: Event) {
-        self.cache.add_path(&event.paths[0]);
+        let recursive_mode = self.recursive_mode(&event.paths[0]);
+
+        self.cache.add_path(&event.paths[0], recursive_mode);
 
         let trackers_match = self
             .rename_event
@@ -378,7 +384,7 @@ impl<T: FileIdCache> DebounceDataInner<T> {
             self.push_rename_event(path, event, time);
         } else {
             // move in
-            self.push_event(event, Instant::now());
+            self.push_event(event, now());
         }
 
         self.rename_event = None;
@@ -519,7 +525,7 @@ pub struct Debouncer<T: Watcher, C: FileIdCache> {
 
 impl<T: Watcher, C: FileIdCache> Debouncer<T, C> {
     /// Stop the debouncer, waits for the event thread to finish.
-    /// May block for the duration of one tick_rate.
+    /// May block for the duration of one `tick_rate`.
     pub fn stop(mut self) {
         self.set_stop();
         if let Some(t) = self.debouncer_thread.take() {
@@ -536,14 +542,60 @@ impl<T: Watcher, C: FileIdCache> Debouncer<T, C> {
         self.stop.store(true, Ordering::Relaxed);
     }
 
-    /// Access to the internally used notify Watcher backend
-    pub fn watcher(&mut self) -> &mut T {
-        &mut self.watcher
+    #[deprecated = "`Debouncer` provides all methods from `Watcher` itself now. Remove `.watcher()` and use those methods directly."]
+    pub fn watcher(&mut self) {}
+
+    #[deprecated = "`Debouncer` now manages root paths automatically. Remove all calls to `add_root` and `remove_root`."]
+    pub fn cache(&mut self) {}
+
+    fn add_root(&mut self, path: impl Into<PathBuf>, recursive_mode: RecursiveMode) {
+        let path = path.into();
+
+        let mut data = self.data.lock().unwrap();
+
+        // skip, if the root has already been added
+        if data.roots.iter().any(|(p, _)| p == &path) {
+            return;
+        }
+
+        data.roots.push((path.clone(), recursive_mode));
+
+        data.cache.add_path(&path, recursive_mode);
     }
 
-    /// Access to the internally used notify Watcher backend
-    pub fn cache(&mut self) -> MappedMutexGuard<C> {
-        MutexGuard::map(self.data.lock(), |data| &mut data.cache)
+    fn remove_root(&mut self, path: impl AsRef<Path>) {
+        let mut data = self.data.lock().unwrap();
+
+        data.roots.retain(|(root, _)| !root.starts_with(&path));
+
+        data.cache.remove_path(path.as_ref());
+    }
+
+    pub fn watch(
+        &mut self,
+        path: impl AsRef<Path>,
+        recursive_mode: RecursiveMode,
+    ) -> notify::Result<()> {
+        self.watcher.watch(path.as_ref(), recursive_mode)?;
+        self.add_root(path.as_ref(), recursive_mode);
+        Ok(())
+    }
+
+    pub fn unwatch(&mut self, path: impl AsRef<Path>) -> notify::Result<()> {
+        self.watcher.unwatch(path.as_ref())?;
+        self.remove_root(path);
+        Ok(())
+    }
+
+    pub fn configure(&mut self, option: notify::Config) -> notify::Result<bool> {
+        self.watcher.configure(option)
+    }
+
+    pub fn kind() -> WatcherKind
+    where
+        Self: Sized,
+    {
+        T::kind()
     }
 }
 
@@ -557,7 +609,7 @@ impl<T: Watcher, C: FileIdCache> Drop for Debouncer<T, C> {
 ///
 /// Timeout is the amount of time after which a debounced event is emitted.
 ///
-/// If tick_rate is None, notify will select a tick rate that is 1/4 of the provided timeout.
+/// If `tick_rate` is `None`, notify will select a tick rate that is 1/4 of the provided timeout.
 pub fn new_debouncer_opt<F: DebounceEventHandler, T: Watcher, C: FileIdCache + Send + 'static>(
     timeout: Duration,
     tick_rate: Option<Duration>,
@@ -599,7 +651,7 @@ pub fn new_debouncer_opt<F: DebounceEventHandler, T: Watcher, C: FileIdCache + S
             let send_data;
             let errors;
             {
-                let mut lock = data_c.lock();
+                let mut lock = data_c.lock().unwrap();
                 send_data = lock.debounced_events();
                 errors = lock.errors();
             }
@@ -614,7 +666,7 @@ pub fn new_debouncer_opt<F: DebounceEventHandler, T: Watcher, C: FileIdCache + S
     let data_c = data.clone();
     let watcher = T::new(
         move |e: Result<Event, Error>| {
-            let mut lock = data_c.lock();
+            let mut lock = data_c.lock().unwrap();
 
             match e {
                 Ok(e) => lock.add_event(e),
@@ -639,31 +691,75 @@ pub fn new_debouncer_opt<F: DebounceEventHandler, T: Watcher, C: FileIdCache + S
 ///
 /// Timeout is the amount of time after which a debounced event is emitted.
 ///
-/// If tick_rate is None, notify will select a tick rate that is 1/4 of the provided timeout.
+/// If `tick_rate` is `None`, notify will select a tick rate that is 1/4 of the provided timeout.
 pub fn new_debouncer<F: DebounceEventHandler>(
     timeout: Duration,
     tick_rate: Option<Duration>,
     event_handler: F,
-) -> Result<Debouncer<RecommendedWatcher, FileIdMap>, Error> {
-    new_debouncer_opt::<F, RecommendedWatcher, FileIdMap>(
+) -> Result<Debouncer<RecommendedWatcher, RecommendedCache>, Error> {
+    new_debouncer_opt::<F, RecommendedWatcher, RecommendedCache>(
         timeout,
         tick_rate,
         event_handler,
-        FileIdMap::new(),
+        RecommendedCache::new(),
         notify::Config::default(),
     )
 }
 
+fn sort_events(events: Vec<DebouncedEvent>) -> Vec<DebouncedEvent> {
+    let mut sorted = Vec::with_capacity(events.len());
+
+    // group events by path
+    let mut events_by_path: HashMap<_, VecDeque<_>> =
+        events.into_iter().fold(HashMap::new(), |mut acc, event| {
+            acc.entry(event.paths.last().cloned().unwrap_or_default())
+                .or_default()
+                .push_back(event);
+            acc
+        });
+
+    // push events for different paths in chronological order and keep the order of events with the same path
+
+    let mut min_time_heap = events_by_path
+        .iter()
+        .map(|(path, events)| Reverse((events[0].time, path.clone())))
+        .collect::<BinaryHeap<_>>();
+
+    while let Some(Reverse((min_time, path))) = min_time_heap.pop() {
+        // unwrap is safe because only paths from `events_by_path` are added to `min_time_heap`
+        // and they are never removed from `events_by_path`.
+        let events = events_by_path.get_mut(&path).unwrap();
+
+        let mut push_next = false;
+
+        while events.front().is_some_and(|event| event.time <= min_time) {
+            // unwrap is safe because `pop_front` mus return some in order to enter the loop
+            let event = events.pop_front().unwrap();
+            sorted.push(event);
+            push_next = true;
+        }
+
+        if push_next {
+            if let Some(event) = events.front() {
+                min_time_heap.push(Reverse((event.time, path)));
+            }
+        }
+    }
+
+    sorted
+}
+
 #[cfg(test)]
 mod tests {
     use std::{fs, path::Path};
 
     use super::*;
 
-    use mock_instant::MockClock;
     use pretty_assertions::assert_eq;
     use rstest::rstest;
+    use tempfile::tempdir;
     use testing::TestCase;
+    use time::MockTime;
 
     #[rstest]
     fn state(
@@ -702,7 +798,9 @@ mod tests {
             "emit_close_events_only_once",
             "emit_modify_event_after_close_event",
             "emit_needs_rescan_event",
-            "read_file_id_without_create_event"
+            "read_file_id_without_create_event",
+            "sort_events_chronologically",
+            "sort_events_with_reordering"
         )]
         file_name: &str,
     ) {
@@ -710,21 +808,25 @@ mod tests {
             fs::read_to_string(Path::new(&format!("./test_cases/{file_name}.hjson"))).unwrap();
         let mut test_case = deser_hjson::from_str::<TestCase>(&file_content).unwrap();
 
-        MockClock::set_time(Duration::default());
-
-        let time = Instant::now();
+        let time = now();
+        MockTime::set_time(time);
 
         let mut state = test_case.state.into_debounce_data_inner(time);
+        state.roots = vec![(PathBuf::from("/"), RecursiveMode::Recursive)];
+
+        let mut prev_event_time = Duration::default();
 
         for event in test_case.events {
+            let event_time = Duration::from_millis(event.time);
             let event = event.into_debounced_event(time, None);
-            MockClock::set_time(event.time - time);
+            MockTime::advance(event_time - prev_event_time);
+            prev_event_time = event_time;
             state.add_event(event.event);
         }
 
         for error in test_case.errors {
-            let e = error.into_notify_error();
-            state.add_error(e);
+            let error = error.into_notify_error();
+            state.add_error(error);
         }
 
         let expected_errors = std::mem::take(&mut test_case.expected.errors);
@@ -760,21 +862,20 @@ mod tests {
             "errors not as expected"
         );
 
-        let backup_time = Instant::now().duration_since(time);
+        let backup_time = now();
         let backup_queues = state.queues.clone();
 
         for (delay, events) in expected_events {
-            MockClock::set_time(backup_time);
+            MockTime::set_time(backup_time);
             state.queues = backup_queues.clone();
 
             match delay.as_str() {
                 "none" => {}
-                "short" => MockClock::advance(Duration::from_millis(10)),
-                "long" => MockClock::advance(Duration::from_millis(100)),
+                "short" => MockTime::advance(Duration::from_millis(10)),
+                "long" => MockTime::advance(Duration::from_millis(100)),
                 _ => {
                     if let Ok(ts) = delay.parse::<u64>() {
-                        let ts = time + Duration::from_millis(ts);
-                        MockClock::set_time(ts - time);
+                        MockTime::set_time(time + Duration::from_millis(ts));
                     }
                 }
             }
@@ -791,4 +892,26 @@ mod tests {
             );
         }
     }
+
+    #[test]
+    fn integration() -> Result<(), Box<dyn std::error::Error>> {
+        let dir = tempdir()?;
+
+        let (tx, rx) = std::sync::mpsc::channel();
+
+        let mut debouncer = new_debouncer(Duration::from_millis(10), None, tx)?;
+
+        debouncer.watch(dir.path(), RecursiveMode::Recursive)?;
+
+        fs::write(dir.path().join("file.txt"), b"Lorem ipsum")?;
+
+        let events = rx
+            .recv_timeout(Duration::from_secs(10))
+            .expect("no events received")
+            .expect("received an error");
+
+        assert!(!events.is_empty(), "received empty event list");
+
+        Ok(())
+    }
 }
